# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.__internal__.distribute namespace.
"""

import sys as _sys

from . import combinations
from . import interim
from . import multi_process_runner
from tensorflow.python.distribute.distribution_strategy_context import variable_sync_on_read_context
from tensorflow.python.distribute.merge_call_interim import strategy_supports_no_merge_call